<?php

/**
 * @package      JCE
 * @copyright    Copyright (C) 2005 - 2013 Ryan Demmer. All rights reserved.
 * @author		Ryan Demmer
 * @license      GNU/GPL
 * JCE is free software. This version may have been modified pursuant
 * to the GNU General Public License, and as distributed it includes or
 * is derivative of works licensed under the GNU General Public License or
 * other free or open source software licenses.
 */
// no direct access
defined('_JEXEC') or die('ERROR_403');

require_once(dirname(__FILE__) . '/s3/s3.php');

class WFS3FileSystem extends WFFileSystem {

    protected $instance;

    /**
     * Constructor activating the default information of the class
     *
     * @access	protected
     */
    public function __construct($config = array()) {
        parent::__construct($config);

        $this->accessKey = $this->getParam('editor.filesystem.s3.accesskey', '');
        $this->secretKey = $this->getParam('editor.filesystem.s3.secretkey', '');
        $this->bucket = $this->getParam('editor.filesystem.s3.bucket', '');
        $this->cname = $this->getParam('editor.filesystem.s3.cname', $this->bucket);
        $this->timeout = (int) $this->getParam('editor.filesystem.s3.timeout', 3600);
        $this->ssl = (int) $this->getParam('editor.filesystem.s3.ssl', 1);
        $this->endpoint = $this->getParam('editor.filesystem.s3.endpoint', 's3.amazonaws.com');

        // create instance
        $this->instance = new S3($this->accessKey, $this->secretKey, (bool) $this->ssl, $this->endpoint);

        //S3::setExceptions(true);

        $elements = array(
            's3_acl' => array(
                'label' => WFText::_('WF_FILESYSTEM_S3_ACL', 'ACL'),
                'options' => array(
                    'private' => WFText::_('WF_FILESYSTEM_S3_ACL_PRIVATE', 'Private'),
                    'public-read' => WFText::_('WF_FILESYSTEM_S3_ACL_PUBLIC_READ', 'Public Read'),
                    'public-read-write' => WFText::_('WF_FILESYSTEM_S3_ACL_PUBLIC_READ_WRITE', 'Public Read/Write'),
                    'authenticated-read' => WFText::_('WF_FILESYSTEM_S3_ACL_AUTHENTICATED_READ', 'Authenticated Read')
                )
            )
        );

        $config = array(
            'local' => false,
            'upload' => array(
                'stream' => false,
                'chunking' => false,
                'unique_filenames' => false,
                'elements' => $elements
            ),
            'folder_new' => array(
                'elements' => $elements
            ),
            'base' => $this->endpoint . '/' . $this->bucket
        );

        $this->setProperties($config);
    }

    /**
     * Get the base directory.
     * @return string base dir
     */
    public function getBaseDir() {
        return $this->getRootDir();
    }

    /**
     * Get the full base url
     * @return string base url
     */
    public function getBaseURL() {
        return $this->getRootDir();
    }

    /**
     * Return the full user directory path. Create if required
     *
     * @param string	The base path
     * @access public
     * @return Full path to folder
     */
    public function getRootDir() {
        static $root;

        if (!isset($root)) {
            $root = parent::getRootDir();
            $wf = WFEditorPlugin::getInstance();

            if (empty($root) && !$wf->getParam('editor.filesystem.s3.allow_root', 0)) {
                $root = 'images';
            }

            if (!empty($root) && !$this->exists($root)) {
                if ($this->folderCreate($root)) {
                    return $root;
                }
            }
        }
        
        return $root;
    }

    /**
     * Count the number of files in a folder
     * @return integer File total
     * @param string $path Absolute path to folder
     */
    public function countItems($relative) {
        return $this->countFiles($relative);
    }

    public function getTotalSize($relative, $recurse = true) {
        $total = 0;

        $list = $this->getItems($relative);

        if (!empty($list)) {
            foreach ($list as $id => $value) {
                if (substr($value['name'], -1) !== "/") {
                    $total += (int) $value['size'];

                    if ($recurse) {
                        $total += $this->getTotalSize($id, $recurse);
                    }
                }
            }
        }

        return $total;
    }

    /**
     * Count the number of files in a given folder
     * @return integer Total number of folders
     * @param string $path Absolute path to folder
     */
    public function countFiles($relative, $recurse = false) {
        $total = 0;

        $list = $this->getItems($relative);

        if (!empty($list)) {
            foreach ($list as $id => $value) {
                // not a folder 
                if (array_key_exists("prefix", $value) === false) {
                    $total++;

                    if ($recurse) {
                        $total += $this->countFiles($id, $recurse);
                    }
                }
            }
        }

        return $total;
    }

    /**
     * Determine whether a key exists
     * @return boolean
     */
    public function exists($path) {
        $path = ltrim($path, "/");
        
        if ($this->is_dir($path)) {
            $path .= '/index.html';
        }
        
        return @$this->instance->getObjectInfo($this->bucket, $path, false);
    }

    public function getFolders($relative, $filter = '') {
        $folders = array();
        $list = $this->getItems($relative);

        if (!empty($list)) {
            natcasesort($list);
            foreach ($list as $id => $value) {

                // check for folder prefix
                if (array_key_exists("prefix", $value)) {

                    // filter
                    if ($filter && preg_match('#' . $filter . '#', $id) === false) {
                        continue;
                    }

                    // get name
                    $item = basename($id);
                    // utf-8 encode
                    $item = WFUtility::isUTF8($item) ? $item : utf8_encode($item);

                    $data = array(
                        'id' => WFUtility::makePath($relative, $item, '/'),
                        'name' => $item,
                        'writable' => true,
                        'type' => 'folders'
                    );

                    $properties = self::getFolderDetails($id, $value);
                    $folders[] = array_merge($data, array('properties' => $properties));
                }
            }
        }

        return $folders;
    }

    private function createURL($path) {
        static $url = array();

        if (empty($url[$path])) {
            $host = $this->ssl ? 'https://' : 'http://';
            $base = WFUtility::makePath($this->endpoint, $this->bucket);
            
            // create public url
            $url[$path] = $host . WFUtility::makePath($base, $path);

            // check ACL
            $acl    = $this->instance->getAccessControlPolicy($this->bucket, $path);
            $public = false;

            if (!empty($acl['acl'])) {
                foreach ($acl['acl'] as $rule) {

                    if ($rule['permission'] === "READ" && (!empty($rule['uri']) && $rule['uri'] === "http://acs.amazonaws.com/groups/global/AllUsers")) {
                        $public = true;
                        break;
                    }
                }
            }
            // get authenticated url if not public
            if (!$public) {
                $url[$path] = $this->instance->getAuthenticatedURL($this->bucket, $path, $this->timeout);
            }
        }

        return $url[$path];
    }

    public function getFiles($relative, $filter) {
        $files = array();
        $list = $this->getItems($relative);

        $x = 1;

        if (!empty($list)) {
            // Sort alphabetically
            natcasesort($list);
            foreach ($list as $key => $values) {
                if ($this->is_file(basename($key))) {

                    if ($filter && preg_match('#' . $filter . '#', $path) === false) {
                        continue;
                    }

                    // create name
                    $item = basename($key);
                    // encode
                    $item = WFUtility::isUTF8($item) ? $item : utf8_encode($item);

                    // create public url
                    $url = $this->createURL($key);

                    // create relative file
                    $id = WFUtility::makePath($relative, $item, '/');

                    // remove leading slash
                    $id = ltrim($id, '/');

                    $data = array(
                        'id' => $id,
                        'url' => $url,
                        'name' => $item,
                        'writable' => true,
                        'type' => 'files'
                    );

                    $properties = self::getFileDetails($key, $x, $values);
                    $files[] = array_merge($data, array('properties' => $properties));

                    $x++;
                }
            }
        }

        return $files;
    }

    protected function getItems($relative) {
        static $items = array();

        $path = WFUtility::makePath($this->getRootDir(), $relative);
        $path = WFUtility::fixPath($path);
        
        $key  = md5($path);
        
        $path = ltrim($path, "/");

        if (empty($items[$key])) {            
            $items[$key] = (array) $this->instance->getBucket($this->bucket, $path, null, null, '/', true);
        }

        return $items[$key];
    }

    /**
     * Get a folders properties
     * 
     * @return array Array of properties
     * @param string $dir Folder relative path
     */
    public function getFolderDetails($dir, $info = array()) {
        if (!empty($info['time'])) {
            $info = $this->instance->getObjectInfo($this->bucket, trim($dir, "/"));
        }
        return array('modified' => strftime($info['time']));
    }

    /**
     * Get a files properties
     * 
     * @return array Array of properties
     * @param string $file File relative path
     */
    public function getFileDetails($file, $count = 1, $info = array()) {
        if (!empty($info['time']) && !empty($info['size'])) {            
            $info = @$this->instance->getObjectInfo($this->bucket, trim($file, "/"));
        }

        $data = array(
            'size' => $info['size'],
            'modified' => strftime($info['time'])
        );

        if (preg_match('/\.(jpeg|jpg|gif|png)/i', $file) && $count < 100) {
            $url = $this->createURL($file);

            $dimensions = '';

            // try getimagesize (may not work on some systems)
            $dim = @getimagesize($url);

            $width = $dim[0];
            $height = $dim[1];

            $image = array(
                'width' => $width,
                'height' => $height,
                'preview' => $url
            );
            return array_merge_recursive($data, $image);
        }

        return $data;
    }

    /**
     * Delete the relative file(s).
     * @param $files the relative path to the file name or comma seperated list of multiple paths.
     * @return string $error on failure.
     */
    public function delete($src, $recursive = false) {
        $path = WFUtility::makePath($this->getBaseDir(), $src);

        // get error class
        $result = new WFFileSystemResult();

        $path = WFUtility::makePath($this->getBaseDir(), $src);
        
        $path = ltrim($path, "/");

        if ($this->is_file($path)) {
            $result->type = 'files';
            $result->state = $this->instance->deleteObject($this->bucket, $path);
        } else {
            $result->type = 'folders';

            if ($recursive == false && $this->countItems($src) > 0) {
                $result->message = JText::sprintf('WF_MANAGER_FOLDER_NOT_EMPTY', basename($path));
            } else {
                $result->state = $this->deleteItems($path);
            }
        }

        return $result;
    }

    protected function deleteItems($src) {
        $items = $this->getItems($src);

        $return = false;

        foreach ($items as $item => $value) {
            // folder
            if (array_key_exists('prefix', $value)) {
                $return = $this->deleteItems($item);
            }
            // file
            $return = $this->instance->deleteObject($this->bucket, $item);
        }

        return $return;
    }

    protected function copyItems($src, $destination, $delete = false) {
        $items = $this->getItems($src);

        if (!empty($items)) {
            foreach ($items as $item => $value) {
                $dest = WFUtility::makePath($destination, basename($item));
                
                $dest = ltrim($dest, "/");

                $state = $this->instance->copyObject($this->bucket, $item, $this->bucket, $dest);
                // remove original
                if ($state && $delete) {
                    $this->instance->deleteObject($this->bucket, $item);
                }
            }
        }

        return true;
    }

    /**
     * Rename a file.
     * @param string $src The relative path of the source file
     * @param string $dest The name of the new file
     * @return string $error
     */
    public function rename($src, $dest) {

        $src = WFUtility::makePath($this->getBaseDir(), rawurldecode($src));
        $dir = dirname($src);
        
        $src = ltrim($src, "/");

        $result = new WFFileSystemResult();

        if ($this->is_file($src)) {
            $ext = JFile::getExt($src);
            $file = $dest . '.' . $ext;
            $dest = WFUtility::makePath($dir, $file);
            
            $dest = ltrim($dest, "/");

            if ($this->exists($dest)) {
                return $result;
            }

            if ($this->instance->copyObject($this->bucket, $src, $this->bucket, $dest)) {
                $result->state = $this->instance->deleteObject($this->bucket, $src);
                $result->path = $path;
            }

            $result->type = 'files';
        } else {
            $dest = WFUtility::makePath($dir, $dest);
            
            $dest = ltrim($dest, "/");

            if ($this->exists($dest)) {
                return $result;
            }

            $this->instance->setExceptions(true);

            // create new folder, copy files
            $result->state = $this->copyItems($src, $dest, true);
            $result->type = 'folders';
        }

        return $result;
    }

    /**
     * Copy a file.
     * @param string $files The relative file or comma seperated list of files
     * @param string $dest The relative path of the destination dir
     * @return string $error on failure
     */
    public function copy($file, $destination, $delete = false) {
        $result = new WFFileSystemResult();

        $src = WFUtility::makePath($this->getBaseDir(), $file);
        $dest = WFUtility::makePath($this->getBaseDir(), WFUtility::makePath($destination, basename($file)));
        
        $src    = ltrim($src, "/");
        $dest   = ltrim($dest, "/");

        // src is a file
        if ($this->is_file($src)) {
            $result->type = 'files';

            $result->state = $this->instance->copyObject($this->bucket, $src, $this->bucket, $dest);

            if ($result->state && $delete) {
                $this->instance->deleteObject($this->bucket, $src);
            }

            $result->path = $dest;
        } else {
            // Folders cannot be copied into themselves as this creates an infinite copy / paste loop	
            if ($file === $destination) {
                $result->state = false;
                $result->message = WFText::_('WF_MANAGER_COPY_INTO_ERROR');

                return $result;
            }
            $result->state = $this->copyItems($src, $dest, $delete);
            $result->type = 'folders';
        }

        return $result;
    }

    /**
     * Copy a file.
     * @param string $files The relative file or comma seperated list of files
     * @param string $dest The relative path of the destination dir
     * @return string $error on failure
     */
    public function move($file, $dest) {
        return $this->copy($file, $dest, true);
    }

    /**
     * New folder base function. A wrapper for the JFolder::create function
     * @param string $folder The folder to create
     * @return boolean true on success
     */
    public function folderCreate($folder, $acl = 'private') {
        $input = '<html><body bgcolor="#FFFFFF"></body></html>';
        return $this->instance->putObject($input, $this->bucket, $folder . '/index.html', $acl);
    }

    /**
     * New folder
     * @param string $dir The base dir
     * @param string $new_dir The folder to be created
     * @return string $error on failure
     */
    public function createFolder($dir, $new) {
        $dir = WFUtility::makePath(rawurldecode($dir), WFUtility::makeSafe($new));
        $dir = WFUtility::makePath($this->getBaseDir(), $dir);
        
        $dir = ltrim($dir, "/");
        
        $acl = JRequest::getWord('s3_acl', 'private');
        $result = new WFFileSystemResult();
        $result->state = $this->folderCreate($dir, $acl);

        return $result;
    }

    public function getDimensions($path) {
        $path = WFUtility::makePath($this->getRootDir(), rawurldecode($path));
        $url = $this->createURL($path);

        $width = 0;
        $height = 0;

        // try getimagesize (may not work on some systems)
        $dim = @getimagesize($url);

        if (!empty($dim)) {
            list($width, $height) = $dim;
        }

        return array(
            'width' => $width,
            'height' => $height
        );
    }

    public function upload($method, $src, $dir, $name, $chunks = 1, $chunk = 0) {
        $path = WFUtility::makePath($this->getRootDir(), rawurldecode($dir));
        $dest = WFUtility::makePath($path, WFUtility::makeSafe($name));
        
        $dest = ltrim($dest, "/");
        
        $result = new WFFileSystemResult();
        $acl = JRequest::getWord('s3_acl', 'private');

        $input = $this->instance->inputFile($src);

        $result->state = $this->instance->putObject($input, $this->bucket, $dest, $acl);

        $result->path = $dest;

        return $result;
    }

    public function read($file) {
        $path = WFUtility::makePath($this->getBaseDir(), rawurldecode($file));
        
        $path = ltrim($path, "/");

        return $this->instance->inputFile($this->bucket, $path);
    }

    public function write($file, $content) {
        $path = WFUtility::makePath($this->getBaseDir(), $file);
        
        $path = ltrim($path, "/");

        return $this->instance->putObject($content, $this->bucket, $path);
    }

    /**
     * Get the source directory of a file path
     */
    public function getSourceDir($path) {        
        if (!empty($path)) {
            // directory path relative to base dir        
            if ($this->is_file($path)) {
                $path = substr(dirname($path), strlen($this->getRootDir()));
            }
        }

        return $path;
    }

    public function isMatch($needle, $haystack) {
        $needle = parse_url($needle, PHP_URL_PATH);
        $haystack = WFUtility::makePath($this->bucket, $haystack);

        return trim($haystack, '/') == ltrim($needle, '/');
    }

    public function pathinfo($path) {
        $path = parse_url($path, PHP_URL_PATH);
        return pathinfo($path);
    }

    public function is_file($path) {
        return preg_match('#\.(' . implode('|', $this->get('filetypes')) . ')$#i', $path);
    }

    public function is_dir($path) {
        return $this->is_file($path) === false;
    }

}
